import importlib.util
import json
import subprocess
from pathlib import Path
from typing import Optional, Type

from deepdiff import DeepDiff
from fastapi import FastAPI, Header
from pydantic import BaseModel, ValidationError
from rich import print
from stringcase import camelcase


class CamelCaseModel(BaseModel):
    class Config:
        alias_generator = camelcase
        allow_population_by_field_name = True


class DataProductDefinition(BaseModel):
    description: Optional[str]
    name: Optional[str]
    request: Type[BaseModel]
    response: Type[BaseModel]
    route_description: Optional[str]
    route_summary: Optional[str]
    summary: str
    requires_authorization: bool = False
    requires_consent: bool = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self.summary:
            if not self.route_description:
                self.route_description = self.summary
            if not self.description:
                self.description = self.summary
            if not self.route_summary:
                self.route_summary = self.summary


def export_openapi_spec(definition: DataProductDefinition) -> dict:
    """
    Given a data product definition, create a FastAPI application and a corresponding
    POST route. Then export its OpenAPI spec
    :param definition: Data product definition
    :return: OpenAPI spec
    """
    app = FastAPI(
        title=definition.summary,
        description=definition.description,
        version="1.0.0",
    )

    if definition.requires_authorization:
        authorization_header_type = str
        authorization_header_default_value = ...
    else:
        authorization_header_type = Optional[str]
        authorization_header_default_value = None

    if definition.requires_consent:
        consent_header_type = str
        consent_header_default_value = ...
        consent_header_description = "Consent token"
    else:
        consent_header_type = Optional[str]
        consent_header_default_value = None
        consent_header_description = "Optional consent token"

    @app.post(
        f"/{definition.name}",
        summary=definition.route_summary,
        description=definition.route_description,
        response_model=definition.response,
    )
    def request(
        params: definition.request,
        x_consent_token: consent_header_type = Header(
            consent_header_default_value,
            description=consent_header_description,
        ),
        authorization: authorization_header_type = Header(
            authorization_header_default_value,
            description='The login token. Value should be "Bearer [token]"',
        ),
        x_authorization_provider: Optional[str] = Header(
            None, description="The bare domain of the system that provided the token."
        ),
    ):
        pass

    openapi = app.openapi()

    for path, data in openapi["paths"].items():
        operation_id = data["post"]["operationId"].removesuffix("_post")
        openapi["paths"][path]["post"]["operationId"] = operation_id

    return openapi


def styled_error(error: str, path: Path) -> str:
    """
    Style error messages to make them clearer and easier to read
    """
    return f"[bold red]{error}[/bold red] in [yellow]{path}[/yellow]:exclamation:"


def convert_data_product_definitions(src: Path, dest: Path) -> bool:
    """
    Browse folder for definitions defined as python files
    and export them to corresponding OpenAPI specs in the output folder
    """

    should_fail_hook = False
    for p in src.glob("**/*.py"):
        spec = importlib.util.spec_from_file_location(name=str(p), location=str(p))
        if not spec.loader:
            raise RuntimeError(f"Failed to import {p} module")
        try:
            module = spec.loader.load_module(str(p))
        except ValidationError as e:
            should_fail_hook = True
            print(styled_error("Validation error", p))
            print(e)
            continue

        try:
            definition: DataProductDefinition = getattr(module, "DEFINITION")
        except AttributeError:
            print(styled_error("Error finding DEFINITION variable", p))
            continue

        # Get definition name based on file path
        definition.name = p.relative_to(src).with_suffix("").as_posix()
        if not definition.route_summary:
            definition.route_summary = definition.name

        openapi = export_openapi_spec(definition)

        out_file = (dest / p.relative_to(src)).with_suffix(".json")

        current_spec = {}
        if out_file.exists():
            current_spec = json.loads(out_file.read_text(encoding="utf-8"))

        # Write resulted JSON only if it's changed to satisfy pre-commit hook
        if DeepDiff(current_spec, openapi, ignore_order=True) != {}:
            print(f"Exporting {out_file}")
            out_file.parent.mkdir(parents=True, exist_ok=True)
            out_file.write_text(
                json.dumps(openapi, indent=2, ensure_ascii=False) + "\n",
                encoding="utf-8",
            )
            run_pre_commit_hooks_on_file(out_file)
            # Hook should fail as we modified the file.
            should_fail_hook = True
        else:
            if file_is_untracked(out_file):
                print(f"Untracked {out_file}")
                should_fail_hook = True
            else:
                print(f"Skipping {out_file}")

    return should_fail_hook


def run_pre_commit_hooks_on_file(file: Path) -> None:
    """
    Run pre-commit hooks on a file.
    """
    subprocess.run(
        [
            "pre-commit",
            "run",
            "--files",
            str(file),
        ],
        capture_output=True,
    )


def file_is_untracked(file: Path) -> bool:
    """
    Check if the file is untracked in git.
    """
    completed_process = subprocess.run(
        ["git", "status", "--short", str(file)],
        check=True,
        capture_output=True,
        encoding="utf-8",
    )
    return completed_process.stdout.startswith("??")



from pathlib import Path

from typer import Argument, Exit, Typer

from converter import convert_data_product_definitions

cli = Typer()


@cli.command()
def convert_definitions(
    src: Path = Argument(
        ...,
        help="Path to python sources of definitions",
        dir_okay=True,
        file_okay=False,
        exists=True,
    ),
    dest: Path = Argument(
        ...,
        help="Path to definitions output",
        dir_okay=True,
        file_okay=False,
        exists=True,
    ),
):
    should_fail_hook = convert_data_product_definitions(src, dest)
    raise Exit(code=int(should_fail_hook))



from pydantic import Field

from converter import CamelCaseModel, DataProductDefinition


class CurrentWeatherMetricRequest(CamelCaseModel):
    lat: float = Field(
        ...,
        title="Latitude",
        description="The latitude coordinate of the desired location",
        ge=-90.0,
        le=90.0,
        example=60.192059,
    )
    lon: float = Field(
        ...,
        title="Longitude",
        description="The longitude coordinate of the desired location",
        ge=-180.0,
        le=180.0,
        example=24.945831,
    )


class CurrentWeatherMetricResponse(CamelCaseModel):
    humidity: float = Field(..., title="Current relative air humidity in %", example=72)
    pressure: float = Field(..., title="Current air pressure in hPa", example=1007)
    rain: bool = Field(
        ..., title="Rain status", description="If it's currently raining or not."
    )
    temp: float = Field(
        ..., title="Current temperature in Celsius", example=17.3, ge=-273.15
    )
    wind_speed: float = Field(..., title="Current wind speed in m/s", example=2.1, ge=0)
    wind_direction: float = Field(
        ...,
        title="Current wind direction in meteorological wind direction degrees",
        ge=0,
        le=360,
        example=220.0,
    )


DEFINITION = DataProductDefinition(
    description="Data Product for current weather with metric units",
    request=CurrentWeatherMetricRequest,
    response=CurrentWeatherMetricResponse,
    route_description="Current weather in metric units",
    summary="Current Weather (Metric)",
    requires_authorization=True,
    requires_consent=True,
)



from typing import List

from pydantic import Field

from converter import CamelCaseModel, DataProductDefinition


class CurrentAirQualityRequest(CamelCaseModel):
    lat: float = Field(
        ...,
        title="Latitude",
        description="The latitude coordinate of the desired location",
        example=60.192059,
        ge=-90,
        le=90,
    )
    lon: float = Field(
        ...,
        title="Longitude",
        description="The longitude coordinate of the desired location",
        example=24.945831,
        ge=-180,
        le=180,
    )


class CurrentAirQualityResponse(CamelCaseModel):
    air_quality_index: int = Field(
        ...,
        title="Air Quality Index",
        description=(
            "Current air quality index.\nRanges:\n0-50 Good;\n51-100 Moderate;\n"
            "101-150 Unhealthy For Sensitive Groups;\n151-200 Unhealthy;\n"
            "201-300 Very Unhealthy;\n301+ Hazardous"
        ),
        ge=0,
        example=30,
    )
    timestamp: str = Field(
        ...,
        title="Timestamp",
        description="Current timestamp in RFC 3339 format",
        example="2020-04-03T13:00:00Z",
    )
    attribution: List[str] = Field(
        ...,
        title="Source Attribution",
        description="List of text to show required credits to data sources",
        example=["Moscow State environmental monitoring"],
    )


DEFINITION = DataProductDefinition(
    description="Data Product for current air quality index",
    request=CurrentAirQualityRequest,
    response=CurrentAirQualityResponse,
    route_description="Current air quality",
    summary="Air Quality Index",
)



import json
from pathlib import Path

import pytest

from converter import (
    CamelCaseModel,
    DataProductDefinition,
    convert_data_product_definitions,
)


def test_air_quality(tmpdir, json_snapshot):
    out_dir = tmpdir.mkdir("output")
    convert_data_product_definitions(Path(__file__).parent / "data", Path(out_dir))

    dest_file = out_dir / "AirQuality" / "Current.json"
    assert dest_file.exists()

    dest_spec = json.loads(dest_file.read_text("utf-8"))
    assert json_snapshot == dest_spec


def test_current_weather_required_headers(tmpdir, json_snapshot):
    out_dir = tmpdir.mkdir("output")
    convert_data_product_definitions(Path(__file__).parent / "data", Path(out_dir))

    dest_file = out_dir / "Weather" / "Current" / "Metric.json"
    assert dest_file.exists()

    dest_spec = json.loads(dest_file.read_text("utf-8"))
    assert json_snapshot == dest_spec


def test_data_product_definition_fallbacks():
    summary = "FooBar summary"

    # Summary as fallback for description and route description
    dpd = DataProductDefinition(
        summary=summary,
        name="Foo/Bar",
        request=CamelCaseModel,
        response=CamelCaseModel,
    )
    assert dpd.description == summary
    assert dpd.summary == summary
    assert dpd.route_description == summary

    # Summary as fallback for route description only
    description = "FooBar description"
    dpd = DataProductDefinition(
        summary=summary,
        description=description,
        name="Foo/Bar",
        request=CamelCaseModel,
        response=CamelCaseModel,
    )
    assert dpd.description == description
    assert dpd.summary == summary
    assert dpd.route_description == summary

    # Summary as fallback for description only
    route_description = "FooBar route description"
    dpd = DataProductDefinition(
        summary=summary,
        route_description=route_description,
        name="Foo/Bar",
        request=CamelCaseModel,
        response=CamelCaseModel,
    )
    assert dpd.description == summary
    assert dpd.summary == summary
    assert dpd.route_description == route_description

    # Missing summary
    with pytest.raises(ValueError):
        DataProductDefinition(
            description=description,
            route_description=route_description,
            name="Foo/Bar",
            request=CamelCaseModel,
            response=CamelCaseModel,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['converter',
 'converter.tests',
 'converter.tests.data.AirQuality',
 'converter.tests.data.Weather.Current']

package_data = \
{'': ['*'], 'converter.tests': ['__snapshots__/test_converter/*']}

install_requires = \
['deepdiff>=6.2.2,<7.0.0',
 'fastapi>=0.88.0,<0.89.0',
 'pydantic[email]>=1.10.4,<2.0.0',
 'stringcase>=1.2.0,<2.0.0',
 'typer[all]>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['converter = converter.cli:cli']}

setup_kwargs = {
    'name': 'data-product-definition-tooling',
    'version': '0.0.9',
    'description': 'Data Product Definition Tooling',
    'long_description': '# Data Product Definitions tooling\n\nTools for managing Data Product definitions\n\n# Installation\n\n```shell\npoetry install\n```\n\n# Usage\n\n```shell\npoetry run converter --help\n\n# run tests\npoetry run invoke test\n\n# release a new version (after bumping it in pyproject.toml)\npoetry run invoke release\n```\n\n## Pre-commit hooks\n\n```yaml\nrepos:\n  - repo: https://github.com/ioxio-dataspace/data-product-definition-tooling\n    rev: main # You probably want to lock this to a specific tag\n    hooks:\n      - id: data-product-definition-converter\n        files: "src/.*py$"\n        args: ["src", "dest"]\n```\n',
    'author': 'Digital Living International Ltd',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ioxio-dataspace/data-product-definition-tooling',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9.0,<4',
}


setup(**setup_kwargs)
